import pandas as pd


def filter_damaged_taxa(df, filter_conditions):
    """Function to filter damaged taxa
    Args:
        df (panda.DataFrame): A dataframe containing metaDMG results
        filter_conditions (dict): A dictionary with filter conditions
        taxonomic_rank (str): Select the taxonomic rank to filter
    Returns:
        pandas.DataFrame: A filtered dataframe containing metaDMG results
    """

    mdmg_results = df.loc[
        (df[list(filter_conditions)] >= pd.Series(filter_conditions)).all(axis=1)
    ].copy()

    return mdmg_results


def load_references(file_path):
    """Function to read a mdmg results file to a pandas dataframe
    Args:
        file_path (str): A file path pointing to a mdmg results file
    Returns:
        pandas.DataFrame: A pandas dataframe containing the mdmg results
    """
    references = pd.read_csv(file_path, sep=",", index_col=None, header=None)
    # check how many columns are in the file
    if len(references.columns) == 2:
        references.columns = ["reference", "reference_renamed"]
        references = dict(zip(references.reference, references.reference_renamed))
    elif len(references.columns) == 1:
        references.columns = ["reference"]
        references["reference_renamed"] = references["reference"]
        references = dict(zip(references.reference, references.reference_renamed))
    else:
        references = None
    return references


def load_reads_and_references(file_path):
    """Function to read a mdmg results file to a pandas dataframe
    Args:
        file_path (str): A file path pointing to a mdmg results file
    Returns:
        pandas.DataFrame: A pandas dataframe containing the mdmg results
    """
    reads_df = pd.read_csv(file_path, sep=",", index_col=None, header=None)
    # check how many columns are in the file
    if len(reads_df.columns) == 3:
        reads_df.columns = ["reference", "read", "reference_renamed"]
        reads = dict(zip(reads_df.read, reads_df.reference))
        references = dict(zip(reads_df.reference, reads_df.reference_renamed))
        return reads, references
    elif len(reads_df.columns) == 2:
        reads_df.columns = ["reference", "read"]
        reads_df["reference_renamed"] = reads_df["reference"]
        reads = dict(zip(reads_df.read, reads_df.reference))
        return reads, None
    else:
        reads = None
        return reads, None


def load_fb_results(file_path):
    """Function to read a filterBAM results file to a pandas dataframe
    Args:
        file_path (str): A file path pointing to a mdmg results file
    Returns:
        pandas.DataFrame: A pandas dataframe containing the mdmg results
    """
    fb_results = pd.read_csv(file_path, sep="\t", index_col=None)
    return fb_results


def filter_fb_references(df, filter_conditions):
    fb_results = df.loc[
        (df[list(filter_conditions)] >= pd.Series(filter_conditions)).all(axis=1)
    ].copy()

    return fb_results



"""
 Copyright (c) 2022 Antonio Fernandez-Guerra

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to
 use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 the Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 """


import logging

from get_read_percid.utils import (
    get_arguments,
)
from get_read_percid.lib import load_references, load_reads_and_references
from get_read_percid.extract import get_read_percid, get_percid
import os
import pysam
import numpy as np
import sys


log = logging.getLogger("my_logger")


def main():
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)s ::: %(asctime)s ::: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    args = get_arguments()
    logging.getLogger("my_logger").setLevel(
        logging.DEBUG if args.debug else logging.INFO
    )

    if args.reference_list and args.read_list:
        log.error("You cannot provide both a reference list and a read list")
        sys.exit(1)

    # Check that rank and taxonomy file are both present
    if args.reference_list:
        log.info("Loading references list...")
        references = load_references(args.reference_list)
        if references is None:
            log.error("The reference list file has an invalid number of columns.")
    else:
        references = None

    if args.read_list:
        log.info("Loading read list...")
        reads, references = load_reads_and_references(args.read_list)
        if references is None or reads is None:
            log.error("The read list file has an invalid number of columns")
    else:
        reads = None

    logging.info("Loading BAM file...")
    save = pysam.set_verbosity(0)
    bam = args.bam
    samfile = pysam.AlignmentFile(bam, "rb", threads=args.threads)

    chr_lengths = []
    for chrom in samfile.references:
        chr_lengths.append(samfile.get_reference_length(chrom))
    max_chr_length = np.max(chr_lengths)

    sorted_bam = bam.replace(".bam", ".id-sorted.bam")
    sorted_bam_index_bai = sorted_bam.replace(".id-sorted.bam", ".id-sorted.bam.bai")
    sorted_bam_index_csi = sorted_bam.replace(".id-sorted.bam", ".id-sorted.bam.csi")
    # Check if BAM files is not sorted by coordinates, sort it by coordinates
    if not samfile.header["HD"]["SO"] == "coordinate":
        log.info("BAM file is not sorted by coordinates, sorting it...")
        pysam.sort(
            "-@", str(args.threads), "-m", str(args.sort_memory), "-o", sorted_bam, bam
        )
        bam = sorted_bam
        samfile = pysam.AlignmentFile(bam, "rb", threads=args.threads)

    if samfile.has_index():
        if os.path.exists(sorted_bam_index_bai):
            os.remove(sorted_bam_index_bai)
        elif os.path.exists(sorted_bam_index_csi):
            os.remove(sorted_bam_index_csi)

    logging.info("Indexing BAM file...")
    if max_chr_length > 536870912:
        logging.info("A reference is longer than 2^29, indexing with csi")
        pysam.index("-c", "-@", str(args.threads), bam)
    else:
        pysam.index(
            "-@", str(args.threads), bam
        )  # Need to reload the samfile after creating index
        log.info("Re-loading BAM file")
        samfile = pysam.AlignmentFile(bam, "rb", threads=args.threads)
    pysam.set_verbosity(save)

    refs_bam = [
        chrom.contig for chrom in samfile.get_index_statistics() if chrom.mapped > 0
    ]
    samfile.close()

    # If we have a taxonomy file, get the references that are in the BAM file
    if references is not None:
        logging.info("Extracting stats and reads for references in list...")
        # Check if references are in the BAM file using set intersection
        refs_tax = set(references.keys()).intersection(set(refs_bam))
        refs_tax = list(refs_tax)
        # check that references in refs_tax are the same than in references
        if not set(refs_tax) == set(references.keys()):
            log.error("Some references in the file are not in the BAM file.")
            exit(1)
        if not refs_tax:
            log.error("No references found in BAM file")
            exit(1)
        reads = get_read_percid(
            bam=bam,
            refs=references,
            reads=reads,
            threads=args.threads,
            prefix=args.prefix,
            chunksize=args.chunk_size,
        )
    else:
        logging.info("Extracting stats and reads for all references...")
        reads = get_percid(
            bam=bam,
            threads=args.threads,
            prefix=args.prefix,
        )

    if os.path.exists(sorted_bam_index_bai):
        os.remove(sorted_bam_index_bai)
    if os.path.exists(sorted_bam_index_csi):
        os.remove(sorted_bam_index_csi)
    if os.path.exists(sorted_bam):
        os.remove(sorted_bam)
    logging.info("Done!")


if __name__ == "__main__":
    main()



import tqdm
import pysam
import logging
from multiprocessing import Pool
from functools import partial
from get_read_percid.utils import is_debug, calc_chunksize, initializer
import numpy as np
import sys
import gzip

# import cProfile as profile
# import pstats

log = logging.getLogger("my_logger")


# function to calculate GC content
def calc_gc_content(seq):
    """Calculate GC content of a sequence

    Args:
        seq (str): DNA sequence

    Returns:
        int: Number of GC content
    """
    gc = seq.count("G") + seq.count("C")
    return gc


def get_percid(bam, prefix=None, threads=1):
    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
    # write results to csv file
    if prefix is not None:
        ofname = f"{prefix}.percid.csv.gz"
    else:
        ofname = "sample.percid.csv.gz"
    # get number of reads in the file
    n_reads = samfile.count(until_eof=True)
    samfile.close()
    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)
    _open = partial(gzip.open, mode="wt") if ofname.endswith(".gz") else open
    with _open(ofname) as f:
        for aln in tqdm.tqdm(
            samfile.fetch(until_eof=True),
            total=n_reads,
            desc="Processing",
            disable=is_debug(),
            unit=" reads",
        ):
            if aln.is_unmapped:
                continue
            ani_read = (1 - ((aln.get_tag("NM") / aln.infer_query_length()))) * 100
            f.write(
                f"{aln.query_name},{aln.reference_name},{aln.infer_query_length()},{ani_read},{calc_gc_content(aln.query_sequence)}\n"
            )
    samfile.close()


def get_ids_and_filter(params, refs, reads, prefix, threads=1):
    bam, references = params

    samfile = pysam.AlignmentFile(bam, "rb", threads=threads)

    # convert the dictionary to an array
    refs_dict = dict(zip(samfile.references, samfile.lengths))
    for reference in references:
        my_array = np.array(list(refs_dict.items()))

        # get the indices of the keys to keep
        keep_indices = np.isin(my_array[:, 0], [reference])

        # use array indexing to get the key-value pairs to keep
        filtered_array = my_array[keep_indices]

        # convert the filtered array back to a dictionary
        refs_dict = dict(filtered_array)
        refs_dict = {refs[x]: y for x, y in refs_dict.items()}

        (ref_names, ref_lengths) = zip(*refs_dict.items())
        ref_lengths = list(ref_lengths)
        # convert reference lengths to integers
        ref_lengths = [int(x) for x in ref_lengths]

        refs_idx = {sys.intern(str(x)): i for i, x in enumerate(ref_names)}

        ref_name = refs[reference]
        # write results to csv file
        if prefix is not None:
            ofname = f"{prefix}___{ref_name}--percid.csv.gz"
            obam_file = f"{prefix}___{ref_name}.bam"
        else:
            ofname = f"{ref_name}--percid.csv.gz"
            obam_file = f"{ref_name}.bam"

        out_bam_file = pysam.AlignmentFile(
            obam_file,
            "wb",
            referencenames=list(ref_names),
            referencelengths=ref_lengths,
            threads=1,
        )

        _open = partial(gzip.open, mode="wt") if ofname.endswith(".gz") else open
        with _open(ofname) as f:
            for aln in samfile.fetch(
                contig=reference, multiple_iterators=False, until_eof=True
            ):
                if aln.is_unmapped:
                    continue
                if reads is None:
                    ani_read = (
                        1 - ((aln.get_tag("NM") / aln.infer_query_length()))
                    ) * 100
                    f.write(
                        f"{aln.query_name},{aln.infer_query_length()},{ani_read},{calc_gc_content(aln.query_sequence)}\n"
                    )
                    aln.reference_id = refs_idx[ref_name]
                    out_bam_file.write(aln)
                else:
                    if aln.query_name in reads:
                        if reads[aln.query_name] == reference:
                            ani_read = (
                                1 - ((aln.get_tag("NM") / aln.infer_query_length()))
                            ) * 100
                            f.write(
                                f"{aln.query_name},{aln.infer_query_length()},{ani_read},{calc_gc_content(aln.query_sequence)}\n"
                            )
                            aln.reference_id = refs_idx[ref_name]
                            out_bam_file.write(aln)
        out_bam_file.close()

    samfile.close()


def get_read_percid(
    bam,
    refs,
    reads,
    prefix=None,
    chunksize=None,
    threads=1,
):
    # prof = profile.Profile()
    # prof.enable()
    ref_list = [key for key in refs]

    if (chunksize is not None) and ((len(ref_list) // chunksize) > threads):
        c_size = chunksize
    else:
        c_size = calc_chunksize(n_workers=threads, len_iterable=len(ref_list), factor=4)

    ref_chunks = [ref_list[i : i + c_size] for i in range(0, len(ref_list), c_size)]

    params = zip([bam] * len(ref_chunks), ref_chunks)

    if is_debug():
        data = list(
            map(
                partial(
                    get_ids_and_filter,
                    refs=refs,
                    reads=reads,
                    prefix=prefix,
                    threads=threads,
                ),
                params,
            )
        )
    else:
        logging.info(
            f"Processing {len(ref_chunks):,} chunks of {c_size:,} references each..."
        )
        p = Pool(
            threads,
            initializer=initializer,
            initargs=([params, refs],),
        )

        data = list(
            tqdm.tqdm(
                p.imap_unordered(
                    partial(
                        get_ids_and_filter,
                        refs=refs,
                        reads=reads,
                        prefix=prefix,
                        threads=threads,
                    ),
                    params,
                    chunksize=1,
                ),
                total=len(ref_chunks),
                leave=False,
                ncols=80,
                desc="References processed",
            )
        )

        p.close()
        p.join()



from setuptools import setup
import versioneer

requirements = [
    "pandas>=1.4.2",
    "tqdm>=4.64.1",
    "pysam>=0.17.0",
    "numpy>=1.21.2",
    "biopython>=1.79",
]

setup(
    setup_requires=[
        # Setuptools 18.0 properly handles Cython extensions.
        "setuptools>=39.1.0",
        "Cython>=0.29.24",
    ],
    name="get-read-percid",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="A simple tool to extract references from BAM files and get read statistics",
    license="MIT",
    author="Antonio Fernandez-Guerra",
    author_email="antonio@metagenomics.eu",
    url="https://github.com/genomewalker/get-read-percid",
    packages=["get_read_percid"],
    entry_points={"console_scripts": ["getRPercId=get_read_percid.__main__:main"]},
    install_requires=requirements,
    keywords="get-read-percid",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)
